"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const getSlothData_1 = require("../utils/getSlothData");
const getProtoData_1 = require("../utils/getProtoData");
/**
 * SlothField decorator is used to mark a specific class property
 * as a document key. This introduces a few behaviors:
 *  - setting the value in the constructor (or directly alongside declaration)
 *    will set its default value
 *  - mutating the value will update it in updatedProps
 *  - accessing the value will first look into updatedProps, then props and then default values
 * @typeparam T the value type
 * @param docKeyName specifies the document key name, default to prop key name
 */
function SlothField(docKeyName) {
    return function (target, key) {
        const docKey = docKeyName || key;
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothField on top of another decorator');
            }
        }
        const data = getProtoData_1.default(target, true);
        data.fields.push({ key, docKey });
        Reflect.deleteProperty(target, key);
        Reflect.defineProperty(target, key, {
            enumerable: true,
            get: function () {
                const { updatedProps, props = {}, defaultProps } = getSlothData_1.default(this);
                if (docKey in updatedProps) {
                    return updatedProps[docKey];
                }
                if (docKey in props) {
                    return props[docKey];
                }
                return defaultProps[docKey];
            },
            set: function (value) {
                const { props, defaultProps, updatedProps } = getSlothData_1.default(this);
                if (!props) {
                    defaultProps[docKey] = value;
                    return;
                }
                if (docKey in defaultProps && value == null) {
                    delete props[docKey];
                    delete updatedProps[docKey];
                    return;
                }
                Object.assign(updatedProps, { [docKey]: value });
            }
        });
    };
}
exports.default = SlothField;
//# sourceMappingURL=SlothField.js.map