"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const getSlothData_1 = require("../utils/getSlothData");
const getProtoData_1 = require("../utils/getProtoData");
/**
 *
 * SlothRel is used to indicate that a specific string property
 * corresponds to another entity identifier. The possible relations are:
 *
 *   - belongsTo: one or several entities belongs to a parent entity
 *   - hasMany: one entity has several other entities; the relation is stored the _id
 *
 * @see [[RelationDescriptor]]
 * @param rel the relation description
 */
function SlothRel(rel) {
    return function (target, key) {
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothRel on top of another decorator');
            }
        }
        const { fields, rels } = getProtoData_1.default(target, true);
        if ('belongsTo' in rel) {
            fields.push({ key, docKey: key });
        }
        rels.push(Object.assign({}, rel, { key }));
        Reflect.deleteProperty(target, key);
        Reflect.defineProperty(target, key, {
            get: function (target = this) {
                if ('hasMany' in rel) {
                    return () => rel.hasMany().withRoot(this._id);
                }
                const { updatedProps, props } = getSlothData_1.default(target);
                if (key in updatedProps) {
                    return updatedProps[key];
                }
                return props[key];
            },
            set: function (value) {
                // Typescript calls this function before class decorator
                // Thus, when assigning default values in constructor we can get it and write it down
                // However this should only happen once to avoid missing bugs
                if (value === '') {
                    return;
                }
                Object.assign(getSlothData_1.default(this).updatedProps, { [key]: value });
            }
        });
    };
}
exports.default = SlothRel;
//# sourceMappingURL=SlothRel.js.map