"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const getSlothData_1 = require("../utils/getSlothData");
const path_1 = require("path");
const getProtoData_1 = require("../utils/getProtoData");
/**
 * The SlothURI decorator describes a class parameter that has no intrisic value
 * but which value depends on the other properties described as a path.
 *
 * The path has a root value, which is a constant and should be either the database name
 * pluralized, or a namespace. For example in library management system, the root could be
 * either `library` or `books`. It is  recommended to use a namespace for relational
 * databases and the database name for orphans.
 *
 * The path components, following the root, should be slugified properties of the entity.
 * If the path needs to include another entity identifier, as often it needs to be in relational
 * database, then the root of the other entity id should be omitted, but path separator (/)
 * should NOT be escaped;, even in URLs. The path components are described using their property names.
 *
 * @param prefix the URI root
 * @param propsKeys key names to pick from the document
 * @typeparam S the document schema
 */
function SlothURI(prefix, ...propsKeys) {
    return (target, key) => {
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothURI on top of another decorator');
            }
        }
        Reflect.deleteProperty(target, key);
        const { uris, fields } = getProtoData_1.default(target, true);
        uris.push({
            name: key,
            prefix,
            propsKeys
        });
        fields.push({ key, docKey: key });
        Reflect.defineProperty(target, key, {
            get: function () {
                const { slug } = getSlothData_1.default(this);
                return path_1.join(prefix, ...propsKeys.map(propKey => {
                    const value = this[propKey];
                    if (!value) {
                        throw new Error(`Key ${propKey} has no value, but is required for the ${key} URI`);
                    }
                    const { rels } = getProtoData_1.default(target);
                    const relation = rels.find(({ key }) => key === propKey);
                    if (relation && 'belongsTo' in relation) {
                        const splat = value.split('/');
                        if (splat.length > 1) {
                            return splat.slice(1).join('/');
                        }
                        throw new Error(`URI '${value}' is invalid'`);
                    }
                    return slug(value.toString());
                }));
            },
            set: function (val) {
                if (val === '') {
                    return;
                }
                throw new Error(`Property ${key} is not writable`);
            }
        });
    };
}
exports.default = SlothURI;
//# sourceMappingURL=SlothURI.js.map