"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const getSlothData_1 = require("../utils/getSlothData");
const getProtoData_1 = require("../utils/getProtoData");
const path_1 = require("path");
const debug_1 = require("debug");
const debug = debug_1.default('slothdb');
const limax = require('limax');
const slug = (text) => limax(text, { separateNumbers: false });
/**
 * Base abstract entity, for all entitoies
 * The generic parameter S is the schema of the document
 * @typeparam S the document schema
 */
class BaseEntity {
    // tslint:disable-next-line:no-empty
    constructor(factory, idOrProps) {
        this._id = '';
        const { name } = getProtoData_1.default(this);
        if (!name) {
            throw new Error('Please use SlothEntity');
        }
        if (typeof idOrProps === 'string') {
            this.sloth = {
                name,
                updatedProps: {},
                defaultProps: {},
                docId: idOrProps,
                factory,
                slug
            };
        }
        else {
            this.sloth = {
                name,
                defaultProps: {},
                updatedProps: {},
                docId: idOrProps._id,
                factory,
                slug
            };
        }
    }
    /**
     * Returns whether this document hhas unsaved updated properties
     */
    isDirty() {
        const { docId, updatedProps } = getSlothData_1.default(this);
        return Object.keys(updatedProps).length > 0 || docId == null;
    }
    /**
     * Returns a list of props following the entity schema
     */
    getProps() {
        const { fields } = getProtoData_1.default(this);
        return fields.reduce((props, { key }) => {
            return Object.assign({}, props, { [key]: this[key] });
        }, {});
    }
    /**
     * Returns a list of props mapped with docKey
     */
    getDocument() {
        const { fields } = getProtoData_1.default(this);
        return fields.reduce((props, { key, docKey }) => {
            return Object.assign({}, props, { [docKey]: this[key] });
        }, {});
    }
    /**
     * Saves document to database. If the document doesn't exist,
     * create it. If it exists, update it. If the _id was changed
     * (due to props changing), remove to old document and create a new one
     *
     * @returns a Promise resolving into document props
     *   - If the document was updated, the _rev prop would be defined
     *     and start with an index greater than 1
     *   - If the document was created, the _rev prop would be defined
     *     and start with 1
     *   - If the document was not updated, because it is not dirty,
     *     then no _rev property is returned
     */
    save() {
        return __awaiter(this, void 0, void 0, function* () {
            debug('save document "%s"', this._id);
            const { fields } = getProtoData_1.default(this, false);
            const doc = this.getDocument();
            if (!this.isDirty()) {
                debug('document "%s" is not dirty, skippping saving', this._id);
                return doc;
            }
            const { factory, name, docId } = getSlothData_1.default(this);
            const db = factory(name);
            try {
                const { _rev } = yield db.get(this._id);
                debug('document "%s" already exists with revision "%s", updating...', this._id, _rev);
                const { rev } = yield db.put(Object.assign({}, doc, { _rev }));
                getSlothData_1.default(this).docId = this._id;
                return Object.assign({}, doc, { _rev: rev });
            }
            catch (err) {
                // Then document was not found
                if (err.name === 'not_found') {
                    if (docId) {
                        debug('document "%s" was renamed to "%s", removing old document', docId, this._id);
                        // We need to delete old doc
                        const originalDoc = yield db.get(docId);
                        yield db.remove(originalDoc);
                        getSlothData_1.default(this).docId = this._id;
                    }
                    const { rev, id } = yield db.put(doc);
                    debug('document "%s" rev "%s" was created', this._id, rev);
                    getSlothData_1.default(this).docId = this._id;
                    return Object.assign({}, doc, { _rev: rev });
                }
                throw err;
            }
        });
    }
    /**
     * Returns whether the specified document exists in database
     * Please note that this does not compare documents
     */
    exists() {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, factory } = getSlothData_1.default(this);
            try {
                yield factory(name).get(this._id);
                return true;
            }
            catch (err) {
                return false;
            }
        });
    }
    /**
     * Remove a document from the database
     * @returns a Promise that resolves into a boolean, true if document was removed,
     *          false if the document doesn't have a docId in its slothdata
     */
    remove() {
        return __awaiter(this, void 0, void 0, function* () {
            const { docId, factory, name } = getSlothData_1.default(this);
            if (!docId) {
                return false;
            }
            const db = factory(name);
            const { _rev } = yield db.get(docId);
            yield db.remove(docId, _rev);
            getSlothData_1.default(this).docId = undefined;
            yield this.removeRelations();
            return true;
        });
    }
    /**
     * @private
     */
    removeRelations() {
        const { rels } = getProtoData_1.default(this, false);
        return Promise.all(rels.map(rel => this.removeEntityRelation(rel))).then(() => {
            return;
        });
    }
    getProp(key) {
        return this[key];
    }
    getRelationDescriptor(keyName) {
        const { rels } = getProtoData_1.default(this);
        const rel = rels.find(rel => {
            return rel.key === keyName;
        });
        return rel;
    }
    getRelationEntity(keyName) {
        const rel = this.getRelationDescriptor(keyName);
        const { factory } = getSlothData_1.default(this);
        if ('belongsTo' in rel) {
            return rel.belongsTo().findById(factory, this.getProp(keyName));
        }
        throw new Error(`Cannot use getRelationEntity on ${keyName} as hasMany relation`);
    }
    removeEntityRelation(rel) {
        return __awaiter(this, void 0, void 0, function* () {
            const { factory, name } = getSlothData_1.default(this);
            if ('belongsTo' in rel && rel.cascade) {
                const relId = this.getProp(rel.key);
                const children = yield factory(name).allDocs({
                    include_docs: false,
                    startkey: path_1.join(relId, '/'),
                    endkey: path_1.join(relId, '/\uffff')
                });
                if (children.rows.length === 0) {
                    try {
                        const parent = yield this.getRelationEntity(rel.key);
                        yield parent.remove();
                    }
                    catch (err) {
                        if (err.message === 'missing') {
                            return;
                        }
                        throw err;
                    }
                }
            }
            if ('hasMany' in rel && rel.cascade !== false) {
                const db = rel.hasMany().withRoot(this._id);
                const children = yield db.findAllDocs(factory);
                yield Promise.all(children.map(child => child.remove()));
            }
        });
    }
}
exports.default = BaseEntity;
//# sourceMappingURL=BaseEntity.js.map