"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const changes_1 = require("./changes");
const getProtoData_1 = require("../utils/getProtoData");
const path_1 = require("path");
const debug_1 = require("debug");
const debug = debug_1.default('slothdb');
/**
 * This represent a Database
 *
 * @typeparam S the database schema
 * @typeparam E the Entity
 * @typeparam V the (optional) view type that defines a list of possible view IDs
 */
class SlothDatabase {
    /**
     * Create a new database instance
     * @param factory the pouch factory to use
     * @param name the database name
     * @param model the model constructor
     * @param root the root name, which is the startKey. I don't recommend it
     */
    constructor(model, root = '') {
        this._subscribers = [];
        this._model = model;
        this._root = root;
        const { name } = getProtoData_1.default(model.prototype);
        /* istanbul ignore if  */
        if (!name) {
            throw new Error('SlothEntity decorator is required');
        }
        this._name = name;
    }
    /**
     * Join URI params provided as specified by SlothURI
     * Useful to recreate document id from URL params
     * Please note that
     * @param props the props
     * @param field
     */
    joinURIParams(props, field = '_id') {
        const idURI = getProtoData_1.default(this._model.prototype).uris.find(({ name }) => name === field);
        if (!idURI) {
            throw new Error(`Field ${field} not found in URIs`);
        }
        return path_1.join(idURI.prefix, ...idURI.propsKeys.map(key => props[key]));
    }
    /**
     * Run a query
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     * @param includeDocs include_docs
     */
    query(factory, view, startKey, endKey = typeof startKey === 'string'
        ? path_1.join(startKey, '\uffff')
        : undefined, includeDocs = false) {
        return factory(this._name)
            .query(view, {
            startkey: startKey,
            endkey: endKey,
            include_docs: includeDocs
        })
            .catch(err => {
            if (err.name === 'not_found') {
                debug(`Design document '%s' is missing, generating views...`, view);
                /* istanbul ignore if  */
                if (this._setupPromise) {
                    this._setupPromise.then(() => {
                        return this.query(factory, view, startKey, endKey, includeDocs);
                    });
                }
                this._setupPromise = this.initSetup(factory);
                return this._setupPromise.then(() => {
                    debug('Created design documents');
                    this._setupPromise = undefined;
                    return this.query(factory, view, startKey, endKey, includeDocs);
                });
            }
            throw err;
        });
    }
    /**
     * Queries and maps docs to Entity objects
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryDocs(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, true).then(({ rows }) => {
            return rows.map(({ doc }) => new this._model(factory, doc));
        });
    }
    /**
     * Queries keys. Returns an array of emitted keys
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeys(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, false).then(({ rows }) => {
            return rows.map(({ key }) => key);
        });
    }
    /**
     * Queries keys/_id map. Returns a map of emitted keys/ID
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeysIDs(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, false).then(({ rows }) => {
            return rows.reduce((acc, { key, id }) => (Object.assign({}, acc, { [key]: id })), {});
        });
    }
    /**
     * Returns a database that will only find entities with _id
     * starting with the root path
     * @param root the root path
     */
    withRoot(root) {
        return new SlothDatabase(this._model, path_1.join(this._root, root));
    }
    /**
     * Fetches all documents IDs for this database and return them
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of string IDs
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllIDs(factory, startKey = this._root || '', endKey = path_1.join(startKey, '\uffff')) {
        const db = factory(this._name);
        return db
            .allDocs({
            include_docs: false,
            startkey: startKey,
            endkey: endKey
        })
            .then(({ rows }) => {
            return rows.map(({ id }) => id);
        });
    }
    /**
     * Fetches all documents for this database and map them with the model
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of entity instances
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllDocs(factory, startKey = this._root || '', endKey = path_1.join(startKey, '\uffff')) {
        const db = factory(this._name);
        return db
            .allDocs({
            include_docs: true,
            startkey: startKey,
            endkey: endKey
        })
            .then(({ rows }) => {
            return rows.map(({ doc }) => this.create(factory, doc));
        });
    }
    /**
     * Fetch a docuemt from the database
     * @param factory the PouchDB factory to use
     * @param id the document identifier to fetch
     * @return a promise resolving with the entity instance
     */
    findById(factory, id) {
        return factory(this._name)
            .get(id)
            .then(res => {
            return new this._model(factory, res);
        });
    }
    /**
     * Create a new model instance
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    create(factory, props) {
        return new this._model(factory, props);
    }
    /**
     * Create a new model instance and save it to database
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    put(factory, props) {
        const doc = new this._model(factory, props);
        return doc.save().then(() => doc);
    }
    /**
     * Subscribes a function to PouchDB changes, so that
     * the function will be called when changes are made
     *
     * @param factory the PouchDB factory
     * @param sub the subscriber function
     * @see [[Subscriber]]
     * @see [[ChangeAction]]
     */
    subscribe(factory, sub) {
        if (!this.getSubscriberFor(factory)) {
            debug('Creating changes ');
            const changes = factory(this._name)
                .changes({
                since: 'now',
                live: true,
                include_docs: true
            })
                .on('change', ({ deleted, doc, id }) => {
                if (deleted || !doc) {
                    return this.dispatch(factory, {
                        type: changes_1.ActionType.REMOVED,
                        payload: { [this._name]: id },
                        meta: {}
                    });
                }
                if (doc._rev.match(/^1-/)) {
                    return this.dispatch(factory, {
                        type: changes_1.ActionType.ADDED,
                        payload: { [this._name]: doc },
                        meta: { revision: doc._rev }
                    });
                }
                return this.dispatch(factory, {
                    type: changes_1.ActionType.CHANGED,
                    payload: { [this._name]: doc },
                    meta: { revision: doc._rev }
                });
            });
            this._subscribers.push({ factory, sub, changes });
            return;
        }
        const { changes } = this.getSubscriberFor(factory);
        this._subscribers.push({ factory, sub, changes });
    }
    /**
     * Unsubscribe a subscriber, so it will not be called anymore
     * Possibly cancel PouchDB changes
     *
     * @param factory The pouchDB factory to unsubscribe from
     * @param sub the subscriber to unsubscribe
     */
    cancel(factory, sub) {
        const index = this._subscribers.findIndex(el => el.factory === factory && el.sub === sub);
        const [{ changes }] = this._subscribers.splice(index, 1);
        if (!this.getSubscriberFor(factory)) {
            changes.cancel();
        }
    }
    /**
     * Creates view documents (if required)
     * @param factory
     */
    initSetup(factory) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.setupViews(factory);
        });
    }
    getSubscriberFor(factory) {
        return this._subscribers.find(el => el.factory === factory);
    }
    dispatch(facto, action) {
        this._subscribers.forEach(({ sub, factory }) => facto === factory && sub(action));
    }
    setupViews(factory) {
        const { views } = getProtoData_1.default(this._model.prototype);
        const db = factory(this._name);
        const promises = views.map(({ name, id, code }) => () => __awaiter(this, void 0, void 0, function* () {
            const views = {};
            let _rev;
            try {
                const doc = (yield db.get(`_design/${id}`));
                if (doc.views[name] && doc.views[name].map === code) {
                    // view already exists and is up-to-date
                    return;
                }
                Object.assign(views, doc.views);
                _rev = doc._rev;
            }
            catch (err) {
                // Do nothing
            }
            yield db.put(Object.assign({}, {
                _id: `_design/${id}`,
                views: Object.assign({}, views, { [name]: {
                        map: code
                    } })
            }, _rev ? { _rev } : {}));
        }));
        return promises.reduce((acc, fn) => {
            return acc.then(() => fn());
        }, Promise.resolve());
    }
}
exports.default = SlothDatabase;
//# sourceMappingURL=SlothDatabase.js.map