/**
 * Extract the ProtoData from a class prototype,
 * possibly creating it if needed
 * @see [[ProtoData]]
 * @param obj the object to extract data from
 * @param createIfNotFound create the protoData if it is undefined
 * @private
 */
function getProtoData(obj, createIfNotFound = false) {
    const wrapped = obj;
    if (!wrapped.__protoData) {
        if (createIfNotFound) {
            wrapped.__protoData = {
                uris: [],
                fields: [],
                rels: [],
                views: []
            };
        }
        else {
            throw new Error(`Object ${wrapped} has no __protoData`);
        }
    }
    return wrapped.__protoData;
}

function mapPropsOrDocToDocument({ fields }, data) {
    if (typeof data === 'string') {
        return {};
    }
    return fields.reduce((props, { key, docKey }) => {
        if (!(key in data) && !(docKey in data)) {
            return props;
        }
        if (key in data && docKey in data && key !== docKey) {
            throw new Error(`Both '${key}' and '${docKey}' exist on ${data}`);
        }
        return Object.assign({}, props, {
            [docKey]: key in data ? data[key] : data[docKey]
        });
    }, {});
}
/**
 * This decorator is used to mark classes that will be an entity, a document
 * This function, by extending the constructor and defining this.sloth property
 * effectively allows the usage of other property decorators
 * @param name The database name for this entity
 * @typeparam S The database schema
 */
function SlothEntity(name) {
    return (constructor) => {
        const data = getProtoData(constructor.prototype, true);
        data.name = name;
        const BaseEntity = constructor;
        return class WrappedEntity extends BaseEntity {
            constructor(factory, idOrProps) {
                super(factory, idOrProps);
                this.sloth.props = mapPropsOrDocToDocument(getProtoData(this), idOrProps);
            }
        };
    };
}

/**
 * Extract sloth data from an entity instance,
 * throwing an exception if it can't be found
 * @param obj the class instance to extract from
 * @typeparam S the entity schema
 * @private
 */
function getSlothData(obj) {
    const wrapped = obj;
    if (!wrapped.sloth) {
        throw new Error(`Class ${wrapped} does not extend SlothEntity`);
    }
    return wrapped.sloth;
}

/**
 * SlothField decorator is used to mark a specific class property
 * as a document key. This introduces a few behaviors:
 *  - setting the value in the constructor (or directly alongside declaration)
 *    will set its default value
 *  - mutating the value will update it in updatedProps
 *  - accessing the value will first look into updatedProps, then props and then default values
 * @typeparam T the value type
 * @param docKeyName specifies the document key name, default to prop key name
 */
function SlothField(docKeyName) {
    return function (target, key) {
        const docKey = docKeyName || key;
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothField on top of another decorator');
            }
        }
        const data = getProtoData(target, true);
        data.fields.push({ key, docKey });
        Reflect.deleteProperty(target, key);
        Reflect.defineProperty(target, key, {
            enumerable: true,
            get: function () {
                const { updatedProps, props = {}, defaultProps } = getSlothData(this);
                if (docKey in updatedProps) {
                    return updatedProps[docKey];
                }
                if (docKey in props) {
                    return props[docKey];
                }
                return defaultProps[docKey];
            },
            set: function (value) {
                const { props, defaultProps, updatedProps } = getSlothData(this);
                if (!props) {
                    defaultProps[docKey] = value;
                    return;
                }
                if (docKey in defaultProps && value == null) {
                    delete props[docKey];
                    delete updatedProps[docKey];
                    return;
                }
                Object.assign(updatedProps, { [docKey]: value });
            }
        });
    };
}

// Copyright Joyent, Inc. and other Node contributors.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to permit
// persons to whom the Software is furnished to do so, subject to the
// following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
// USE OR OTHER DEALINGS IN THE SOFTWARE.

// resolves . and .. elements in a path array with directory names there
// must be no slashes, empty elements, or device names (c:\) in the array
// (so also no leading and trailing slashes - it does not distinguish
// relative and absolute paths)
function normalizeArray(parts, allowAboveRoot) {
  // if the path tries to go above the root, `up` ends up > 0
  var up = 0;
  for (var i = parts.length - 1; i >= 0; i--) {
    var last = parts[i];
    if (last === '.') {
      parts.splice(i, 1);
    } else if (last === '..') {
      parts.splice(i, 1);
      up++;
    } else if (up) {
      parts.splice(i, 1);
      up--;
    }
  }

  // if the path is allowed to go above the root, restore leading ..s
  if (allowAboveRoot) {
    for (; up--; up) {
      parts.unshift('..');
    }
  }

  return parts;
}

// path.normalize(path)
// posix version
function normalize(path) {
  var isPathAbsolute = isAbsolute(path),
      trailingSlash = substr(path, -1) === '/';

  // Normalize the path
  path = normalizeArray(filter(path.split('/'), function(p) {
    return !!p;
  }), !isPathAbsolute).join('/');

  if (!path && !isPathAbsolute) {
    path = '.';
  }
  if (path && trailingSlash) {
    path += '/';
  }

  return (isPathAbsolute ? '/' : '') + path;
}
// posix version
function isAbsolute(path) {
  return path.charAt(0) === '/';
}

// posix version
function join() {
  var paths = Array.prototype.slice.call(arguments, 0);
  return normalize(filter(paths, function(p, index) {
    if (typeof p !== 'string') {
      throw new TypeError('Arguments to path.join must be strings');
    }
    return p;
  }).join('/'));
}
function filter (xs, f) {
    if (xs.filter) return xs.filter(f);
    var res = [];
    for (var i = 0; i < xs.length; i++) {
        if (f(xs[i], i, xs)) res.push(xs[i]);
    }
    return res;
}

// String.prototype.substr - negative index don't work in IE8
var substr = 'ab'.substr(-1) === 'b' ?
    function (str, start, len) { return str.substr(start, len) } :
    function (str, start, len) {
        if (start < 0) start = str.length + start;
        return str.substr(start, len);
    }
;

/**
 * The SlothURI decorator describes a class parameter that has no intrisic value
 * but which value depends on the other properties described as a path.
 *
 * The path has a root value, which is a constant and should be either the database name
 * pluralized, or a namespace. For example in library management system, the root could be
 * either `library` or `books`. It is  recommended to use a namespace for relational
 * databases and the database name for orphans.
 *
 * The path components, following the root, should be slugified properties of the entity.
 * If the path needs to include another entity identifier, as often it needs to be in relational
 * database, then the root of the other entity id should be omitted, but path separator (/)
 * should NOT be escaped;, even in URLs. The path components are described using their property names.
 *
 * @param prefix the URI root
 * @param propsKeys key names to pick from the document
 * @typeparam S the document schema
 */
function SlothURI(prefix, ...propsKeys) {
    return (target, key) => {
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothURI on top of another decorator');
            }
        }
        Reflect.deleteProperty(target, key);
        const { uris, fields } = getProtoData(target, true);
        uris.push({
            name: key,
            prefix,
            propsKeys
        });
        fields.push({ key, docKey: key });
        Reflect.defineProperty(target, key, {
            get: function () {
                const { slug } = getSlothData(this);
                return join(prefix, ...propsKeys.map(propKey => {
                    const value = this[propKey];
                    if (!value) {
                        throw new Error(`Key ${propKey} has no value, but is required for the ${key} URI`);
                    }
                    const { rels } = getProtoData(target);
                    const relation = rels.find(({ key }) => key === propKey);
                    if (relation && 'belongsTo' in relation) {
                        const splat = value.split('/');
                        if (splat.length > 1) {
                            return splat.slice(1).join('/');
                        }
                        throw new Error(`URI '${value}' is invalid'`);
                    }
                    return slug(value.toString());
                }));
            },
            set: function (val) {
                if (val === '') {
                    return;
                }
                throw new Error(`Property ${key} is not writable`);
            }
        });
    };
}

/**
 *
 * SlothRel is used to indicate that a specific string property
 * corresponds to another entity identifier. The possible relations are:
 *
 *   - belongsTo: one or several entities belongs to a parent entity
 *   - hasMany: one entity has several other entities; the relation is stored the _id
 *
 * @see [[RelationDescriptor]]
 * @param rel the relation description
 */
function SlothRel(rel) {
    return function (target, key) {
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (desc.get || desc.set) {
                throw new Error('Cannot apply SlothRel on top of another decorator');
            }
        }
        const { fields, rels } = getProtoData(target, true);
        if ('belongsTo' in rel) {
            fields.push({ key, docKey: key });
        }
        rels.push(Object.assign({}, rel, { key }));
        Reflect.deleteProperty(target, key);
        Reflect.defineProperty(target, key, {
            get: function (target = this) {
                if ('hasMany' in rel) {
                    return () => rel.hasMany().withRoot(this._id);
                }
                const { updatedProps, props } = getSlothData(target);
                if (key in updatedProps) {
                    return updatedProps[key];
                }
                return props[key];
            },
            set: function (value) {
                // Typescript calls this function before class decorator
                // Thus, when assigning default values in constructor we can get it and write it down
                // However this should only happen once to avoid missing bugs
                if (value === '') {
                    return;
                }
                Object.assign(getSlothData(this).updatedProps, { [key]: value });
            }
        });
    };
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

/**
 * Helpers.
 */

var s = 1000;
var m = s * 60;
var h = m * 60;
var d = h * 24;
var y = d * 365.25;

/**
 * Parse or format the given `val`.
 *
 * Options:
 *
 *  - `long` verbose formatting [false]
 *
 * @param {String|Number} val
 * @param {Object} [options]
 * @throws {Error} throw an error if val is not a non-empty string or a number
 * @return {String|Number}
 * @api public
 */

var ms = function(val, options) {
  options = options || {};
  var type = typeof val;
  if (type === 'string' && val.length > 0) {
    return parse(val);
  } else if (type === 'number' && isNaN(val) === false) {
    return options.long ? fmtLong(val) : fmtShort(val);
  }
  throw new Error(
    'val is not a non-empty string or a valid number. val=' +
      JSON.stringify(val)
  );
};

/**
 * Parse the given `str` and return milliseconds.
 *
 * @param {String} str
 * @return {Number}
 * @api private
 */

function parse(str) {
  str = String(str);
  if (str.length > 100) {
    return;
  }
  var match = /^((?:\d+)?\.?\d+) *(milliseconds?|msecs?|ms|seconds?|secs?|s|minutes?|mins?|m|hours?|hrs?|h|days?|d|years?|yrs?|y)?$/i.exec(
    str
  );
  if (!match) {
    return;
  }
  var n = parseFloat(match[1]);
  var type = (match[2] || 'ms').toLowerCase();
  switch (type) {
    case 'years':
    case 'year':
    case 'yrs':
    case 'yr':
    case 'y':
      return n * y;
    case 'days':
    case 'day':
    case 'd':
      return n * d;
    case 'hours':
    case 'hour':
    case 'hrs':
    case 'hr':
    case 'h':
      return n * h;
    case 'minutes':
    case 'minute':
    case 'mins':
    case 'min':
    case 'm':
      return n * m;
    case 'seconds':
    case 'second':
    case 'secs':
    case 'sec':
    case 's':
      return n * s;
    case 'milliseconds':
    case 'millisecond':
    case 'msecs':
    case 'msec':
    case 'ms':
      return n;
    default:
      return undefined;
  }
}

/**
 * Short format for `ms`.
 *
 * @param {Number} ms
 * @return {String}
 * @api private
 */

function fmtShort(ms) {
  if (ms >= d) {
    return Math.round(ms / d) + 'd';
  }
  if (ms >= h) {
    return Math.round(ms / h) + 'h';
  }
  if (ms >= m) {
    return Math.round(ms / m) + 'm';
  }
  if (ms >= s) {
    return Math.round(ms / s) + 's';
  }
  return ms + 'ms';
}

/**
 * Long format for `ms`.
 *
 * @param {Number} ms
 * @return {String}
 * @api private
 */

function fmtLong(ms) {
  return plural(ms, d, 'day') ||
    plural(ms, h, 'hour') ||
    plural(ms, m, 'minute') ||
    plural(ms, s, 'second') ||
    ms + ' ms';
}

/**
 * Pluralization helper.
 */

function plural(ms, n, name) {
  if (ms < n) {
    return;
  }
  if (ms < n * 1.5) {
    return Math.floor(ms / n) + ' ' + name;
  }
  return Math.ceil(ms / n) + ' ' + name + 's';
}

var ms$1 = /*#__PURE__*/Object.freeze({
  default: ms,
  __moduleExports: ms
});

var require$$0 = ( ms$1 && ms ) || ms$1;

var debug = createCommonjsModule(function (module, exports) {
/**
 * This is the common logic for both the Node.js and web browser
 * implementations of `debug()`.
 *
 * Expose `debug()` as the module.
 */

exports = module.exports = createDebug.debug = createDebug['default'] = createDebug;
exports.coerce = coerce;
exports.disable = disable;
exports.enable = enable;
exports.enabled = enabled;
exports.humanize = require$$0;

/**
 * Active `debug` instances.
 */
exports.instances = [];

/**
 * The currently active debug mode names, and names to skip.
 */

exports.names = [];
exports.skips = [];

/**
 * Map of special "%n" handling functions, for the debug "format" argument.
 *
 * Valid key names are a single, lower or upper-case letter, i.e. "n" and "N".
 */

exports.formatters = {};

/**
 * Select a color.
 * @param {String} namespace
 * @return {Number}
 * @api private
 */

function selectColor(namespace) {
  var hash = 0, i;

  for (i in namespace) {
    hash  = ((hash << 5) - hash) + namespace.charCodeAt(i);
    hash |= 0; // Convert to 32bit integer
  }

  return exports.colors[Math.abs(hash) % exports.colors.length];
}

/**
 * Create a debugger with the given `namespace`.
 *
 * @param {String} namespace
 * @return {Function}
 * @api public
 */

function createDebug(namespace) {

  var prevTime;

  function debug() {
    // disabled?
    if (!debug.enabled) return;

    var self = debug;

    // set `diff` timestamp
    var curr = +new Date();
    var ms = curr - (prevTime || curr);
    self.diff = ms;
    self.prev = prevTime;
    self.curr = curr;
    prevTime = curr;

    // turn the `arguments` into a proper Array
    var args = new Array(arguments.length);
    for (var i = 0; i < args.length; i++) {
      args[i] = arguments[i];
    }

    args[0] = exports.coerce(args[0]);

    if ('string' !== typeof args[0]) {
      // anything else let's inspect with %O
      args.unshift('%O');
    }

    // apply any `formatters` transformations
    var index = 0;
    args[0] = args[0].replace(/%([a-zA-Z%])/g, function(match, format) {
      // if we encounter an escaped % then don't increase the array index
      if (match === '%%') return match;
      index++;
      var formatter = exports.formatters[format];
      if ('function' === typeof formatter) {
        var val = args[index];
        match = formatter.call(self, val);

        // now we need to remove `args[index]` since it's inlined in the `format`
        args.splice(index, 1);
        index--;
      }
      return match;
    });

    // apply env-specific formatting (colors, etc.)
    exports.formatArgs.call(self, args);

    var logFn = debug.log || exports.log || console.log.bind(console);
    logFn.apply(self, args);
  }

  debug.namespace = namespace;
  debug.enabled = exports.enabled(namespace);
  debug.useColors = exports.useColors();
  debug.color = selectColor(namespace);
  debug.destroy = destroy;

  // env-specific initialization logic for debug instances
  if ('function' === typeof exports.init) {
    exports.init(debug);
  }

  exports.instances.push(debug);

  return debug;
}

function destroy () {
  var index = exports.instances.indexOf(this);
  if (index !== -1) {
    exports.instances.splice(index, 1);
    return true;
  } else {
    return false;
  }
}

/**
 * Enables a debug mode by namespaces. This can include modes
 * separated by a colon and wildcards.
 *
 * @param {String} namespaces
 * @api public
 */

function enable(namespaces) {
  exports.save(namespaces);

  exports.names = [];
  exports.skips = [];

  var i;
  var split = (typeof namespaces === 'string' ? namespaces : '').split(/[\s,]+/);
  var len = split.length;

  for (i = 0; i < len; i++) {
    if (!split[i]) continue; // ignore empty strings
    namespaces = split[i].replace(/\*/g, '.*?');
    if (namespaces[0] === '-') {
      exports.skips.push(new RegExp('^' + namespaces.substr(1) + '$'));
    } else {
      exports.names.push(new RegExp('^' + namespaces + '$'));
    }
  }

  for (i = 0; i < exports.instances.length; i++) {
    var instance = exports.instances[i];
    instance.enabled = exports.enabled(instance.namespace);
  }
}

/**
 * Disable debug output.
 *
 * @api public
 */

function disable() {
  exports.enable('');
}

/**
 * Returns true if the given mode name is enabled, false otherwise.
 *
 * @param {String} name
 * @return {Boolean}
 * @api public
 */

function enabled(name) {
  if (name[name.length - 1] === '*') {
    return true;
  }
  var i, len;
  for (i = 0, len = exports.skips.length; i < len; i++) {
    if (exports.skips[i].test(name)) {
      return false;
    }
  }
  for (i = 0, len = exports.names.length; i < len; i++) {
    if (exports.names[i].test(name)) {
      return true;
    }
  }
  return false;
}

/**
 * Coerce `val`.
 *
 * @param {Mixed} val
 * @return {Mixed}
 * @api private
 */

function coerce(val) {
  if (val instanceof Error) return val.stack || val.message;
  return val;
}
});
var debug_1 = debug.coerce;
var debug_2 = debug.disable;
var debug_3 = debug.enable;
var debug_4 = debug.enabled;
var debug_5 = debug.humanize;
var debug_6 = debug.instances;
var debug_7 = debug.names;
var debug_8 = debug.skips;
var debug_9 = debug.formatters;

var debug$1 = /*#__PURE__*/Object.freeze({
  default: debug,
  __moduleExports: debug,
  coerce: debug_1,
  disable: debug_2,
  enable: debug_3,
  enabled: debug_4,
  humanize: debug_5,
  instances: debug_6,
  names: debug_7,
  skips: debug_8,
  formatters: debug_9
});

var require$$0$1 = ( debug$1 && debug ) || debug$1;

var browser = createCommonjsModule(function (module, exports) {
/**
 * This is the web browser implementation of `debug()`.
 *
 * Expose `debug()` as the module.
 */

exports = module.exports = require$$0$1;
exports.log = log;
exports.formatArgs = formatArgs;
exports.save = save;
exports.load = load;
exports.useColors = useColors;
exports.storage = 'undefined' != typeof chrome
               && 'undefined' != typeof chrome.storage
                  ? chrome.storage.local
                  : localstorage();

/**
 * Colors.
 */

exports.colors = [
  '#0000CC', '#0000FF', '#0033CC', '#0033FF', '#0066CC', '#0066FF', '#0099CC',
  '#0099FF', '#00CC00', '#00CC33', '#00CC66', '#00CC99', '#00CCCC', '#00CCFF',
  '#3300CC', '#3300FF', '#3333CC', '#3333FF', '#3366CC', '#3366FF', '#3399CC',
  '#3399FF', '#33CC00', '#33CC33', '#33CC66', '#33CC99', '#33CCCC', '#33CCFF',
  '#6600CC', '#6600FF', '#6633CC', '#6633FF', '#66CC00', '#66CC33', '#9900CC',
  '#9900FF', '#9933CC', '#9933FF', '#99CC00', '#99CC33', '#CC0000', '#CC0033',
  '#CC0066', '#CC0099', '#CC00CC', '#CC00FF', '#CC3300', '#CC3333', '#CC3366',
  '#CC3399', '#CC33CC', '#CC33FF', '#CC6600', '#CC6633', '#CC9900', '#CC9933',
  '#CCCC00', '#CCCC33', '#FF0000', '#FF0033', '#FF0066', '#FF0099', '#FF00CC',
  '#FF00FF', '#FF3300', '#FF3333', '#FF3366', '#FF3399', '#FF33CC', '#FF33FF',
  '#FF6600', '#FF6633', '#FF9900', '#FF9933', '#FFCC00', '#FFCC33'
];

/**
 * Currently only WebKit-based Web Inspectors, Firefox >= v31,
 * and the Firebug extension (any Firefox version) are known
 * to support "%c" CSS customizations.
 *
 * TODO: add a `localStorage` variable to explicitly enable/disable colors
 */

function useColors() {
  // NB: In an Electron preload script, document will be defined but not fully
  // initialized. Since we know we're in Chrome, we'll just detect this case
  // explicitly
  if (typeof window !== 'undefined' && window.process && window.process.type === 'renderer') {
    return true;
  }

  // Internet Explorer and Edge do not support colors.
  if (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/(edge|trident)\/(\d+)/)) {
    return false;
  }

  // is webkit? http://stackoverflow.com/a/16459606/376773
  // document is undefined in react-native: https://github.com/facebook/react-native/pull/1632
  return (typeof document !== 'undefined' && document.documentElement && document.documentElement.style && document.documentElement.style.WebkitAppearance) ||
    // is firebug? http://stackoverflow.com/a/398120/376773
    (typeof window !== 'undefined' && window.console && (window.console.firebug || (window.console.exception && window.console.table))) ||
    // is firefox >= v31?
    // https://developer.mozilla.org/en-US/docs/Tools/Web_Console#Styling_messages
    (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/firefox\/(\d+)/) && parseInt(RegExp.$1, 10) >= 31) ||
    // double check webkit in userAgent just in case we are in a worker
    (typeof navigator !== 'undefined' && navigator.userAgent && navigator.userAgent.toLowerCase().match(/applewebkit\/(\d+)/));
}

/**
 * Map %j to `JSON.stringify()`, since no Web Inspectors do that by default.
 */

exports.formatters.j = function(v) {
  try {
    return JSON.stringify(v);
  } catch (err) {
    return '[UnexpectedJSONParseError]: ' + err.message;
  }
};


/**
 * Colorize log arguments if enabled.
 *
 * @api public
 */

function formatArgs(args) {
  var useColors = this.useColors;

  args[0] = (useColors ? '%c' : '')
    + this.namespace
    + (useColors ? ' %c' : ' ')
    + args[0]
    + (useColors ? '%c ' : ' ')
    + '+' + exports.humanize(this.diff);

  if (!useColors) return;

  var c = 'color: ' + this.color;
  args.splice(1, 0, c, 'color: inherit');

  // the final "%c" is somewhat tricky, because there could be other
  // arguments passed either before or after the %c, so we need to
  // figure out the correct index to insert the CSS into
  var index = 0;
  var lastC = 0;
  args[0].replace(/%[a-zA-Z%]/g, function(match) {
    if ('%%' === match) return;
    index++;
    if ('%c' === match) {
      // we only are interested in the *last* %c
      // (the user may have provided their own)
      lastC = index;
    }
  });

  args.splice(lastC, 0, c);
}

/**
 * Invokes `console.log()` when available.
 * No-op when `console.log` is not a "function".
 *
 * @api public
 */

function log() {
  // this hackery is required for IE8/9, where
  // the `console.log` function doesn't have 'apply'
  return 'object' === typeof console
    && console.log
    && Function.prototype.apply.call(console.log, console, arguments);
}

/**
 * Save `namespaces`.
 *
 * @param {String} namespaces
 * @api private
 */

function save(namespaces) {
  try {
    if (null == namespaces) {
      exports.storage.removeItem('debug');
    } else {
      exports.storage.debug = namespaces;
    }
  } catch(e) {}
}

/**
 * Load `namespaces`.
 *
 * @return {String} returns the previously persisted debug modes
 * @api private
 */

function load() {
  var r;
  try {
    r = exports.storage.debug;
  } catch(e) {}

  // If debug isn't set in LS, and we're in Electron, try to load $DEBUG
  if (!r && typeof process !== 'undefined' && 'env' in process) {
    r = process.env.DEBUG;
  }

  return r;
}

/**
 * Enable namespaces listed in `localStorage.debug` initially.
 */

exports.enable(load());

/**
 * Localstorage attempts to return the localstorage.
 *
 * This is necessary because safari throws
 * when a user disables cookies/localstorage
 * and you attempt to access it.
 *
 * @return {LocalStorage}
 * @api private
 */

function localstorage() {
  try {
    return window.localStorage;
  } catch (e) {}
}
});
var browser_1 = browser.log;
var browser_2 = browser.formatArgs;
var browser_3 = browser.save;
var browser_4 = browser.load;
var browser_5 = browser.useColors;
var browser_6 = browser.storage;
var browser_7 = browser.colors;

var browser$1 = /*#__PURE__*/Object.freeze({
  default: browser,
  __moduleExports: browser,
  log: browser_1,
  formatArgs: browser_2,
  save: browser_3,
  load: browser_4,
  useColors: browser_5,
  storage: browser_6,
  colors: browser_7
});

// MIT lisence
// from https://github.com/substack/tty-browserify/blob/1ba769a6429d242f36226538835b4034bf6b7886/index.js

function isatty() {
  return false;
}

function ReadStream() {
  throw new Error('tty.ReadStream is not implemented');
}

function WriteStream() {
  throw new Error('tty.ReadStream is not implemented');
}

var tty = {
  isatty: isatty,
  ReadStream: ReadStream,
  WriteStream: WriteStream
}

var tty$1 = /*#__PURE__*/Object.freeze({
  isatty: isatty,
  ReadStream: ReadStream,
  WriteStream: WriteStream,
  default: tty
});

// shim for using process in browser
// based off https://github.com/defunctzombie/node-process/blob/master/browser.js

function defaultSetTimout() {
    throw new Error('setTimeout has not been defined');
}
function defaultClearTimeout () {
    throw new Error('clearTimeout has not been defined');
}
var cachedSetTimeout = defaultSetTimout;
var cachedClearTimeout = defaultClearTimeout;
if (typeof global.setTimeout === 'function') {
    cachedSetTimeout = setTimeout;
}
if (typeof global.clearTimeout === 'function') {
    cachedClearTimeout = clearTimeout;
}

function runTimeout(fun) {
    if (cachedSetTimeout === setTimeout) {
        //normal enviroments in sane situations
        return setTimeout(fun, 0);
    }
    // if setTimeout wasn't available but was latter defined
    if ((cachedSetTimeout === defaultSetTimout || !cachedSetTimeout) && setTimeout) {
        cachedSetTimeout = setTimeout;
        return setTimeout(fun, 0);
    }
    try {
        // when when somebody has screwed with setTimeout but no I.E. maddness
        return cachedSetTimeout(fun, 0);
    } catch(e){
        try {
            // When we are in I.E. but the script has been evaled so I.E. doesn't trust the global object when called normally
            return cachedSetTimeout.call(null, fun, 0);
        } catch(e){
            // same as above but when it's a version of I.E. that must have the global object for 'this', hopfully our context correct otherwise it will throw a global error
            return cachedSetTimeout.call(this, fun, 0);
        }
    }


}
function runClearTimeout(marker) {
    if (cachedClearTimeout === clearTimeout) {
        //normal enviroments in sane situations
        return clearTimeout(marker);
    }
    // if clearTimeout wasn't available but was latter defined
    if ((cachedClearTimeout === defaultClearTimeout || !cachedClearTimeout) && clearTimeout) {
        cachedClearTimeout = clearTimeout;
        return clearTimeout(marker);
    }
    try {
        // when when somebody has screwed with setTimeout but no I.E. maddness
        return cachedClearTimeout(marker);
    } catch (e){
        try {
            // When we are in I.E. but the script has been evaled so I.E. doesn't  trust the global object when called normally
            return cachedClearTimeout.call(null, marker);
        } catch (e){
            // same as above but when it's a version of I.E. that must have the global object for 'this', hopfully our context correct otherwise it will throw a global error.
            // Some versions of I.E. have different rules for clearTimeout vs setTimeout
            return cachedClearTimeout.call(this, marker);
        }
    }



}
var queue = [];
var draining = false;
var currentQueue;
var queueIndex = -1;

function cleanUpNextTick() {
    if (!draining || !currentQueue) {
        return;
    }
    draining = false;
    if (currentQueue.length) {
        queue = currentQueue.concat(queue);
    } else {
        queueIndex = -1;
    }
    if (queue.length) {
        drainQueue();
    }
}

function drainQueue() {
    if (draining) {
        return;
    }
    var timeout = runTimeout(cleanUpNextTick);
    draining = true;

    var len = queue.length;
    while(len) {
        currentQueue = queue;
        queue = [];
        while (++queueIndex < len) {
            if (currentQueue) {
                currentQueue[queueIndex].run();
            }
        }
        queueIndex = -1;
        len = queue.length;
    }
    currentQueue = null;
    draining = false;
    runClearTimeout(timeout);
}
function nextTick(fun) {
    var args = new Array(arguments.length - 1);
    if (arguments.length > 1) {
        for (var i = 1; i < arguments.length; i++) {
            args[i - 1] = arguments[i];
        }
    }
    queue.push(new Item(fun, args));
    if (queue.length === 1 && !draining) {
        runTimeout(drainQueue);
    }
}
// v8 likes predictible objects
function Item(fun, array) {
    this.fun = fun;
    this.array = array;
}
Item.prototype.run = function () {
    this.fun.apply(null, this.array);
};
var title = 'browser';
var platform = 'browser';
var browser$2 = true;
var env = {};
var argv = [];
var version = ''; // empty string to avoid regexp issues
var versions = {};
var release = {};
var config = {};

function noop() {}

var on = noop;
var addListener = noop;
var once = noop;
var off = noop;
var removeListener = noop;
var removeAllListeners = noop;
var emit = noop;

function binding(name) {
    throw new Error('process.binding is not supported');
}

function cwd () { return '/' }
function chdir (dir) {
    throw new Error('process.chdir is not supported');
}function umask() { return 0; }

// from https://github.com/kumavis/browser-process-hrtime/blob/master/index.js
var performance = global.performance || {};
var performanceNow =
  performance.now        ||
  performance.mozNow     ||
  performance.msNow      ||
  performance.oNow       ||
  performance.webkitNow  ||
  function(){ return (new Date()).getTime() };

// generate timestamp or delta
// see http://nodejs.org/api/process.html#process_process_hrtime
function hrtime(previousTimestamp){
  var clocktime = performanceNow.call(performance)*1e-3;
  var seconds = Math.floor(clocktime);
  var nanoseconds = Math.floor((clocktime%1)*1e9);
  if (previousTimestamp) {
    seconds = seconds - previousTimestamp[0];
    nanoseconds = nanoseconds - previousTimestamp[1];
    if (nanoseconds<0) {
      seconds--;
      nanoseconds += 1e9;
    }
  }
  return [seconds,nanoseconds]
}

var startTime = new Date();
function uptime() {
  var currentTime = new Date();
  var dif = currentTime - startTime;
  return dif / 1000;
}

var process$1 = {
  nextTick: nextTick,
  title: title,
  browser: browser$2,
  env: env,
  argv: argv,
  version: version,
  versions: versions,
  on: on,
  addListener: addListener,
  once: once,
  off: off,
  removeListener: removeListener,
  removeAllListeners: removeAllListeners,
  emit: emit,
  binding: binding,
  cwd: cwd,
  chdir: chdir,
  umask: umask,
  hrtime: hrtime,
  platform: platform,
  release: release,
  config: config,
  uptime: uptime
};

var inherits;
if (typeof Object.create === 'function'){
  inherits = function inherits(ctor, superCtor) {
    // implementation from standard node.js 'util' module
    ctor.super_ = superCtor;
    ctor.prototype = Object.create(superCtor.prototype, {
      constructor: {
        value: ctor,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
  };
} else {
  inherits = function inherits(ctor, superCtor) {
    ctor.super_ = superCtor;
    var TempCtor = function () {};
    TempCtor.prototype = superCtor.prototype;
    ctor.prototype = new TempCtor();
    ctor.prototype.constructor = ctor;
  };
}
var inherits$1 = inherits;

// Copyright Joyent, Inc. and other Node contributors.
var formatRegExp = /%[sdj%]/g;
function format(f) {
  if (!isString(f)) {
    var objects = [];
    for (var i = 0; i < arguments.length; i++) {
      objects.push(inspect(arguments[i]));
    }
    return objects.join(' ');
  }

  var i = 1;
  var args = arguments;
  var len = args.length;
  var str = String(f).replace(formatRegExp, function(x) {
    if (x === '%%') return '%';
    if (i >= len) return x;
    switch (x) {
      case '%s': return String(args[i++]);
      case '%d': return Number(args[i++]);
      case '%j':
        try {
          return JSON.stringify(args[i++]);
        } catch (_) {
          return '[Circular]';
        }
      default:
        return x;
    }
  });
  for (var x = args[i]; i < len; x = args[++i]) {
    if (isNull(x) || !isObject(x)) {
      str += ' ' + x;
    } else {
      str += ' ' + inspect(x);
    }
  }
  return str;
}

// Mark that a method should not be used.
// Returns a modified function which warns once by default.
// If --no-deprecation is set, then it is a no-op.
function deprecate(fn, msg) {
  // Allow for deprecating things in the process of starting up.
  if (isUndefined(global.process)) {
    return function() {
      return deprecate(fn, msg).apply(this, arguments);
    };
  }

  if (process$1.noDeprecation === true) {
    return fn;
  }

  var warned = false;
  function deprecated() {
    if (!warned) {
      if (process$1.throwDeprecation) {
        throw new Error(msg);
      } else if (process$1.traceDeprecation) {
        console.trace(msg);
      } else {
        console.error(msg);
      }
      warned = true;
    }
    return fn.apply(this, arguments);
  }

  return deprecated;
}

var debugs = {};
var debugEnviron;
function debuglog(set) {
  if (isUndefined(debugEnviron))
    debugEnviron = process$1.env.NODE_DEBUG || '';
  set = set.toUpperCase();
  if (!debugs[set]) {
    if (new RegExp('\\b' + set + '\\b', 'i').test(debugEnviron)) {
      var pid = 0;
      debugs[set] = function() {
        var msg = format.apply(null, arguments);
        console.error('%s %d: %s', set, pid, msg);
      };
    } else {
      debugs[set] = function() {};
    }
  }
  return debugs[set];
}

/**
 * Echos the value of a value. Trys to print the value out
 * in the best way possible given the different types.
 *
 * @param {Object} obj The object to print out.
 * @param {Object} opts Optional options object that alters the output.
 */
/* legacy: obj, showHidden, depth, colors*/
function inspect(obj, opts) {
  // default options
  var ctx = {
    seen: [],
    stylize: stylizeNoColor
  };
  // legacy...
  if (arguments.length >= 3) ctx.depth = arguments[2];
  if (arguments.length >= 4) ctx.colors = arguments[3];
  if (isBoolean(opts)) {
    // legacy...
    ctx.showHidden = opts;
  } else if (opts) {
    // got an "options" object
    _extend(ctx, opts);
  }
  // set default options
  if (isUndefined(ctx.showHidden)) ctx.showHidden = false;
  if (isUndefined(ctx.depth)) ctx.depth = 2;
  if (isUndefined(ctx.colors)) ctx.colors = false;
  if (isUndefined(ctx.customInspect)) ctx.customInspect = true;
  if (ctx.colors) ctx.stylize = stylizeWithColor;
  return formatValue(ctx, obj, ctx.depth);
}

// http://en.wikipedia.org/wiki/ANSI_escape_code#graphics
inspect.colors = {
  'bold' : [1, 22],
  'italic' : [3, 23],
  'underline' : [4, 24],
  'inverse' : [7, 27],
  'white' : [37, 39],
  'grey' : [90, 39],
  'black' : [30, 39],
  'blue' : [34, 39],
  'cyan' : [36, 39],
  'green' : [32, 39],
  'magenta' : [35, 39],
  'red' : [31, 39],
  'yellow' : [33, 39]
};

// Don't use 'blue' not visible on cmd.exe
inspect.styles = {
  'special': 'cyan',
  'number': 'yellow',
  'boolean': 'yellow',
  'undefined': 'grey',
  'null': 'bold',
  'string': 'green',
  'date': 'magenta',
  // "name": intentionally not styling
  'regexp': 'red'
};


function stylizeWithColor(str, styleType) {
  var style = inspect.styles[styleType];

  if (style) {
    return '\u001b[' + inspect.colors[style][0] + 'm' + str +
           '\u001b[' + inspect.colors[style][1] + 'm';
  } else {
    return str;
  }
}


function stylizeNoColor(str, styleType) {
  return str;
}


function arrayToHash(array) {
  var hash = {};

  array.forEach(function(val, idx) {
    hash[val] = true;
  });

  return hash;
}


function formatValue(ctx, value, recurseTimes) {
  // Provide a hook for user-specified inspect functions.
  // Check that value is an object with an inspect function on it
  if (ctx.customInspect &&
      value &&
      isFunction(value.inspect) &&
      // Filter out the util module, it's inspect function is special
      value.inspect !== inspect &&
      // Also filter out any prototype objects using the circular check.
      !(value.constructor && value.constructor.prototype === value)) {
    var ret = value.inspect(recurseTimes, ctx);
    if (!isString(ret)) {
      ret = formatValue(ctx, ret, recurseTimes);
    }
    return ret;
  }

  // Primitive types cannot have properties
  var primitive = formatPrimitive(ctx, value);
  if (primitive) {
    return primitive;
  }

  // Look up the keys of the object.
  var keys = Object.keys(value);
  var visibleKeys = arrayToHash(keys);

  if (ctx.showHidden) {
    keys = Object.getOwnPropertyNames(value);
  }

  // IE doesn't make error fields non-enumerable
  // http://msdn.microsoft.com/en-us/library/ie/dww52sbt(v=vs.94).aspx
  if (isError(value)
      && (keys.indexOf('message') >= 0 || keys.indexOf('description') >= 0)) {
    return formatError(value);
  }

  // Some type of object without properties can be shortcutted.
  if (keys.length === 0) {
    if (isFunction(value)) {
      var name = value.name ? ': ' + value.name : '';
      return ctx.stylize('[Function' + name + ']', 'special');
    }
    if (isRegExp(value)) {
      return ctx.stylize(RegExp.prototype.toString.call(value), 'regexp');
    }
    if (isDate(value)) {
      return ctx.stylize(Date.prototype.toString.call(value), 'date');
    }
    if (isError(value)) {
      return formatError(value);
    }
  }

  var base = '', array = false, braces = ['{', '}'];

  // Make Array say that they are Array
  if (isArray(value)) {
    array = true;
    braces = ['[', ']'];
  }

  // Make functions say that they are functions
  if (isFunction(value)) {
    var n = value.name ? ': ' + value.name : '';
    base = ' [Function' + n + ']';
  }

  // Make RegExps say that they are RegExps
  if (isRegExp(value)) {
    base = ' ' + RegExp.prototype.toString.call(value);
  }

  // Make dates with properties first say the date
  if (isDate(value)) {
    base = ' ' + Date.prototype.toUTCString.call(value);
  }

  // Make error with message first say the error
  if (isError(value)) {
    base = ' ' + formatError(value);
  }

  if (keys.length === 0 && (!array || value.length == 0)) {
    return braces[0] + base + braces[1];
  }

  if (recurseTimes < 0) {
    if (isRegExp(value)) {
      return ctx.stylize(RegExp.prototype.toString.call(value), 'regexp');
    } else {
      return ctx.stylize('[Object]', 'special');
    }
  }

  ctx.seen.push(value);

  var output;
  if (array) {
    output = formatArray(ctx, value, recurseTimes, visibleKeys, keys);
  } else {
    output = keys.map(function(key) {
      return formatProperty(ctx, value, recurseTimes, visibleKeys, key, array);
    });
  }

  ctx.seen.pop();

  return reduceToSingleString(output, base, braces);
}


function formatPrimitive(ctx, value) {
  if (isUndefined(value))
    return ctx.stylize('undefined', 'undefined');
  if (isString(value)) {
    var simple = '\'' + JSON.stringify(value).replace(/^"|"$/g, '')
                                             .replace(/'/g, "\\'")
                                             .replace(/\\"/g, '"') + '\'';
    return ctx.stylize(simple, 'string');
  }
  if (isNumber(value))
    return ctx.stylize('' + value, 'number');
  if (isBoolean(value))
    return ctx.stylize('' + value, 'boolean');
  // For some reason typeof null is "object", so special case here.
  if (isNull(value))
    return ctx.stylize('null', 'null');
}


function formatError(value) {
  return '[' + Error.prototype.toString.call(value) + ']';
}


function formatArray(ctx, value, recurseTimes, visibleKeys, keys) {
  var output = [];
  for (var i = 0, l = value.length; i < l; ++i) {
    if (hasOwnProperty(value, String(i))) {
      output.push(formatProperty(ctx, value, recurseTimes, visibleKeys,
          String(i), true));
    } else {
      output.push('');
    }
  }
  keys.forEach(function(key) {
    if (!key.match(/^\d+$/)) {
      output.push(formatProperty(ctx, value, recurseTimes, visibleKeys,
          key, true));
    }
  });
  return output;
}


function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
  var name, str, desc;
  desc = Object.getOwnPropertyDescriptor(value, key) || { value: value[key] };
  if (desc.get) {
    if (desc.set) {
      str = ctx.stylize('[Getter/Setter]', 'special');
    } else {
      str = ctx.stylize('[Getter]', 'special');
    }
  } else {
    if (desc.set) {
      str = ctx.stylize('[Setter]', 'special');
    }
  }
  if (!hasOwnProperty(visibleKeys, key)) {
    name = '[' + key + ']';
  }
  if (!str) {
    if (ctx.seen.indexOf(desc.value) < 0) {
      if (isNull(recurseTimes)) {
        str = formatValue(ctx, desc.value, null);
      } else {
        str = formatValue(ctx, desc.value, recurseTimes - 1);
      }
      if (str.indexOf('\n') > -1) {
        if (array) {
          str = str.split('\n').map(function(line) {
            return '  ' + line;
          }).join('\n').substr(2);
        } else {
          str = '\n' + str.split('\n').map(function(line) {
            return '   ' + line;
          }).join('\n');
        }
      }
    } else {
      str = ctx.stylize('[Circular]', 'special');
    }
  }
  if (isUndefined(name)) {
    if (array && key.match(/^\d+$/)) {
      return str;
    }
    name = JSON.stringify('' + key);
    if (name.match(/^"([a-zA-Z_][a-zA-Z_0-9]*)"$/)) {
      name = name.substr(1, name.length - 2);
      name = ctx.stylize(name, 'name');
    } else {
      name = name.replace(/'/g, "\\'")
                 .replace(/\\"/g, '"')
                 .replace(/(^"|"$)/g, "'");
      name = ctx.stylize(name, 'string');
    }
  }

  return name + ': ' + str;
}


function reduceToSingleString(output, base, braces) {
  var numLinesEst = 0;
  var length = output.reduce(function(prev, cur) {
    numLinesEst++;
    if (cur.indexOf('\n') >= 0) numLinesEst++;
    return prev + cur.replace(/\u001b\[\d\d?m/g, '').length + 1;
  }, 0);

  if (length > 60) {
    return braces[0] +
           (base === '' ? '' : base + '\n ') +
           ' ' +
           output.join(',\n  ') +
           ' ' +
           braces[1];
  }

  return braces[0] + base + ' ' + output.join(', ') + ' ' + braces[1];
}


// NOTE: These type checking functions intentionally don't use `instanceof`
// because it is fragile and can be easily faked with `Object.create()`.
function isArray(ar) {
  return Array.isArray(ar);
}

function isBoolean(arg) {
  return typeof arg === 'boolean';
}

function isNull(arg) {
  return arg === null;
}

function isNullOrUndefined(arg) {
  return arg == null;
}

function isNumber(arg) {
  return typeof arg === 'number';
}

function isString(arg) {
  return typeof arg === 'string';
}

function isSymbol(arg) {
  return typeof arg === 'symbol';
}

function isUndefined(arg) {
  return arg === void 0;
}

function isRegExp(re) {
  return isObject(re) && objectToString(re) === '[object RegExp]';
}

function isObject(arg) {
  return typeof arg === 'object' && arg !== null;
}

function isDate(d) {
  return isObject(d) && objectToString(d) === '[object Date]';
}

function isError(e) {
  return isObject(e) &&
      (objectToString(e) === '[object Error]' || e instanceof Error);
}

function isFunction(arg) {
  return typeof arg === 'function';
}

function isPrimitive(arg) {
  return arg === null ||
         typeof arg === 'boolean' ||
         typeof arg === 'number' ||
         typeof arg === 'string' ||
         typeof arg === 'symbol' ||  // ES6 symbol
         typeof arg === 'undefined';
}

function isBuffer(maybeBuf) {
  return Buffer.isBuffer(maybeBuf);
}

function objectToString(o) {
  return Object.prototype.toString.call(o);
}


function pad(n) {
  return n < 10 ? '0' + n.toString(10) : n.toString(10);
}


var months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
              'Oct', 'Nov', 'Dec'];

// 26 Feb 16:19:34
function timestamp() {
  var d = new Date();
  var time = [pad(d.getHours()),
              pad(d.getMinutes()),
              pad(d.getSeconds())].join(':');
  return [d.getDate(), months[d.getMonth()], time].join(' ');
}


// log is just a thin wrapper to console.log that prepends a timestamp
function log() {
  console.log('%s - %s', timestamp(), format.apply(null, arguments));
}

function _extend(origin, add) {
  // Don't do anything if add isn't an object
  if (!add || !isObject(add)) return origin;

  var keys = Object.keys(add);
  var i = keys.length;
  while (i--) {
    origin[keys[i]] = add[keys[i]];
  }
  return origin;
}
function hasOwnProperty(obj, prop) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
}

var util = {
  inherits: inherits$1,
  _extend: _extend,
  log: log,
  isBuffer: isBuffer,
  isPrimitive: isPrimitive,
  isFunction: isFunction,
  isError: isError,
  isDate: isDate,
  isObject: isObject,
  isRegExp: isRegExp,
  isUndefined: isUndefined,
  isSymbol: isSymbol,
  isString: isString,
  isNumber: isNumber,
  isNullOrUndefined: isNullOrUndefined,
  isNull: isNull,
  isBoolean: isBoolean,
  isArray: isArray,
  inspect: inspect,
  deprecate: deprecate,
  format: format,
  debuglog: debuglog
}

var util$1 = /*#__PURE__*/Object.freeze({
  format: format,
  deprecate: deprecate,
  debuglog: debuglog,
  inspect: inspect,
  isArray: isArray,
  isBoolean: isBoolean,
  isNull: isNull,
  isNullOrUndefined: isNullOrUndefined,
  isNumber: isNumber,
  isString: isString,
  isSymbol: isSymbol,
  isUndefined: isUndefined,
  isRegExp: isRegExp,
  isObject: isObject,
  isDate: isDate,
  isError: isError,
  isFunction: isFunction,
  isPrimitive: isPrimitive,
  isBuffer: isBuffer,
  log: log,
  inherits: inherits$1,
  _extend: _extend,
  default: util
});

var argv$1 = process.argv;

var terminator = argv$1.indexOf('--');
var hasFlag = function (flag) {
	flag = '--' + flag;
	var pos = argv$1.indexOf(flag);
	return pos !== -1 && (terminator !== -1 ? pos < terminator : true);
};

var supportsColor = (function () {
	if ('FORCE_COLOR' in process.env) {
		return true;
	}

	if (hasFlag('no-color') ||
		hasFlag('no-colors') ||
		hasFlag('color=false')) {
		return false;
	}

	if (hasFlag('color') ||
		hasFlag('colors') ||
		hasFlag('color=true') ||
		hasFlag('color=always')) {
		return true;
	}

	if (process.stdout && !process.stdout.isTTY) {
		return false;
	}

	if (process.platform === 'win32') {
		return true;
	}

	if ('COLORTERM' in process.env) {
		return true;
	}

	if (process.env.TERM === 'dumb') {
		return false;
	}

	if (/^screen|^xterm|^vt100|color|ansi|cygwin|linux/i.test(process.env.TERM)) {
		return true;
	}

	return false;
})();

var supportsColor$1 = /*#__PURE__*/Object.freeze({
  default: supportsColor,
  __moduleExports: supportsColor
});

var tty$2 = ( tty$1 && tty ) || tty$1;

var util$2 = ( util$1 && util ) || util$1;

var require$$1 = ( supportsColor$1 && supportsColor ) || supportsColor$1;

var node = createCommonjsModule(function (module, exports) {
/**
 * Module dependencies.
 */




/**
 * This is the Node.js implementation of `debug()`.
 *
 * Expose `debug()` as the module.
 */

exports = module.exports = require$$0$1;
exports.init = init;
exports.log = log;
exports.formatArgs = formatArgs;
exports.save = save;
exports.load = load;
exports.useColors = useColors;

/**
 * Colors.
 */

exports.colors = [ 6, 2, 3, 4, 5, 1 ];

try {
  var supportsColor = require$$1;
  if (supportsColor && supportsColor.level >= 2) {
    exports.colors = [
      20, 21, 26, 27, 32, 33, 38, 39, 40, 41, 42, 43, 44, 45, 56, 57, 62, 63, 68,
      69, 74, 75, 76, 77, 78, 79, 80, 81, 92, 93, 98, 99, 112, 113, 128, 129, 134,
      135, 148, 149, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171,
      172, 173, 178, 179, 184, 185, 196, 197, 198, 199, 200, 201, 202, 203, 204,
      205, 206, 207, 208, 209, 214, 215, 220, 221
    ];
  }
} catch (err) {
  // swallow - we only care if `supports-color` is available; it doesn't have to be.
}

/**
 * Build up the default `inspectOpts` object from the environment variables.
 *
 *   $ DEBUG_COLORS=no DEBUG_DEPTH=10 DEBUG_SHOW_HIDDEN=enabled node script.js
 */

exports.inspectOpts = Object.keys(process.env).filter(function (key) {
  return /^debug_/i.test(key);
}).reduce(function (obj, key) {
  // camel-case
  var prop = key
    .substring(6)
    .toLowerCase()
    .replace(/_([a-z])/g, function (_, k) { return k.toUpperCase() });

  // coerce string value into JS value
  var val = process.env[key];
  if (/^(yes|on|true|enabled)$/i.test(val)) val = true;
  else if (/^(no|off|false|disabled)$/i.test(val)) val = false;
  else if (val === 'null') val = null;
  else val = Number(val);

  obj[prop] = val;
  return obj;
}, {});

/**
 * Is stdout a TTY? Colored output is enabled when `true`.
 */

function useColors() {
  return 'colors' in exports.inspectOpts
    ? Boolean(exports.inspectOpts.colors)
    : tty$2.isatty(process.stderr.fd);
}

/**
 * Map %o to `util.inspect()`, all on a single line.
 */

exports.formatters.o = function(v) {
  this.inspectOpts.colors = this.useColors;
  return util$2.inspect(v, this.inspectOpts)
    .split('\n').map(function(str) {
      return str.trim()
    }).join(' ');
};

/**
 * Map %o to `util.inspect()`, allowing multiple lines if needed.
 */

exports.formatters.O = function(v) {
  this.inspectOpts.colors = this.useColors;
  return util$2.inspect(v, this.inspectOpts);
};

/**
 * Adds ANSI color escape codes if enabled.
 *
 * @api public
 */

function formatArgs(args) {
  var name = this.namespace;
  var useColors = this.useColors;

  if (useColors) {
    var c = this.color;
    var colorCode = '\u001b[3' + (c < 8 ? c : '8;5;' + c);
    var prefix = '  ' + colorCode + ';1m' + name + ' ' + '\u001b[0m';

    args[0] = prefix + args[0].split('\n').join('\n' + prefix);
    args.push(colorCode + 'm+' + exports.humanize(this.diff) + '\u001b[0m');
  } else {
    args[0] = getDate() + name + ' ' + args[0];
  }
}

function getDate() {
  if (exports.inspectOpts.hideDate) {
    return '';
  } else {
    return new Date().toISOString() + ' ';
  }
}

/**
 * Invokes `util.format()` with the specified arguments and writes to stderr.
 */

function log() {
  return process.stderr.write(util$2.format.apply(util$2, arguments) + '\n');
}

/**
 * Save `namespaces`.
 *
 * @param {String} namespaces
 * @api private
 */

function save(namespaces) {
  if (null == namespaces) {
    // If you set a process.env field to null or undefined, it gets cast to the
    // string 'null' or 'undefined'. Just delete instead.
    delete process.env.DEBUG;
  } else {
    process.env.DEBUG = namespaces;
  }
}

/**
 * Load `namespaces`.
 *
 * @return {String} returns the previously persisted debug modes
 * @api private
 */

function load() {
  return process.env.DEBUG;
}

/**
 * Init logic for `debug` instances.
 *
 * Create a new `inspectOpts` object in case `useColors` is set
 * differently for a particular `debug` instance.
 */

function init (debug) {
  debug.inspectOpts = {};

  var keys = Object.keys(exports.inspectOpts);
  for (var i = 0; i < keys.length; i++) {
    debug.inspectOpts[keys[i]] = exports.inspectOpts[keys[i]];
  }
}

/**
 * Enable namespaces listed in `process.env.DEBUG` initially.
 */

exports.enable(load());
});
var node_1 = node.init;
var node_2 = node.log;
var node_3 = node.formatArgs;
var node_4 = node.save;
var node_5 = node.load;
var node_6 = node.useColors;
var node_7 = node.colors;
var node_8 = node.inspectOpts;

var node$1 = /*#__PURE__*/Object.freeze({
  default: node,
  __moduleExports: node,
  init: node_1,
  log: node_2,
  formatArgs: node_3,
  save: node_4,
  load: node_5,
  useColors: node_6,
  colors: node_7,
  inspectOpts: node_8
});

var require$$0$2 = ( browser$1 && browser ) || browser$1;

var require$$1$1 = ( node$1 && node ) || node$1;

var src = createCommonjsModule(function (module) {
/**
 * Detect Electron renderer process, which is node, but we should
 * treat as a browser.
 */

if (typeof process === 'undefined' || process.type === 'renderer') {
  module.exports = require$$0$2;
} else {
  module.exports = require$$1$1;
}
});

const debug$2 = src('slothdb');
const limax = require('limax');
const slug = (text) => limax(text, { separateNumbers: false });
/**
 * Base abstract entity, for all entitoies
 * The generic parameter S is the schema of the document
 * @typeparam S the document schema
 */
class BaseEntity {
    // tslint:disable-next-line:no-empty
    constructor(factory, idOrProps) {
        this._id = '';
        const { name } = getProtoData(this);
        if (!name) {
            throw new Error('Please use SlothEntity');
        }
        if (typeof idOrProps === 'string') {
            this.sloth = {
                name,
                updatedProps: {},
                defaultProps: {},
                docId: idOrProps,
                factory,
                slug
            };
        }
        else {
            this.sloth = {
                name,
                defaultProps: {},
                updatedProps: {},
                docId: idOrProps._id,
                factory,
                slug
            };
        }
    }
    /**
     * Returns whether this document hhas unsaved updated properties
     */
    isDirty() {
        const { docId, updatedProps } = getSlothData(this);
        return Object.keys(updatedProps).length > 0 || docId == null;
    }
    /**
     * Returns a list of props following the entity schema
     */
    getProps() {
        const { fields } = getProtoData(this);
        return fields.reduce((props, { key }) => {
            return Object.assign({}, props, { [key]: this[key] });
        }, {});
    }
    /**
     * Returns a list of props mapped with docKey
     */
    getDocument() {
        const { fields } = getProtoData(this);
        return fields.reduce((props, { key, docKey }) => {
            return Object.assign({}, props, { [docKey]: this[key] });
        }, {});
    }
    /**
     * Saves document to database. If the document doesn't exist,
     * create it. If it exists, update it. If the _id was changed
     * (due to props changing), remove to old document and create a new one
     *
     * @returns a Promise resolving into document props
     *   - If the document was updated, the _rev prop would be defined
     *     and start with an index greater than 1
     *   - If the document was created, the _rev prop would be defined
     *     and start with 1
     *   - If the document was not updated, because it is not dirty,
     *     then no _rev property is returned
     */
    save() {
        return __awaiter(this, void 0, void 0, function* () {
            debug$2('save document "%s"', this._id);
            const { fields } = getProtoData(this, false);
            const doc = this.getDocument();
            if (!this.isDirty()) {
                debug$2('document "%s" is not dirty, skippping saving', this._id);
                return doc;
            }
            const { factory, name, docId } = getSlothData(this);
            const db = factory(name);
            try {
                const { _rev } = yield db.get(this._id);
                debug$2('document "%s" already exists with revision "%s", updating...', this._id, _rev);
                const { rev } = yield db.put(Object.assign({}, doc, { _rev }));
                getSlothData(this).docId = this._id;
                return Object.assign({}, doc, { _rev: rev });
            }
            catch (err) {
                // Then document was not found
                if (err.name === 'not_found') {
                    if (docId) {
                        debug$2('document "%s" was renamed to "%s", removing old document', docId, this._id);
                        // We need to delete old doc
                        const originalDoc = yield db.get(docId);
                        yield db.remove(originalDoc);
                        getSlothData(this).docId = this._id;
                    }
                    const { rev, id } = yield db.put(doc);
                    debug$2('document "%s" rev "%s" was created', this._id, rev);
                    getSlothData(this).docId = this._id;
                    return Object.assign({}, doc, { _rev: rev });
                }
                throw err;
            }
        });
    }
    /**
     * Returns whether the specified document exists in database
     * Please note that this does not compare documents
     */
    exists() {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, factory } = getSlothData(this);
            try {
                yield factory(name).get(this._id);
                return true;
            }
            catch (err) {
                return false;
            }
        });
    }
    /**
     * Remove a document from the database
     * @returns a Promise that resolves into a boolean, true if document was removed,
     *          false if the document doesn't have a docId in its slothdata
     */
    remove() {
        return __awaiter(this, void 0, void 0, function* () {
            const { docId, factory, name } = getSlothData(this);
            if (!docId) {
                return false;
            }
            const db = factory(name);
            const { _rev } = yield db.get(docId);
            yield db.remove(docId, _rev);
            getSlothData(this).docId = undefined;
            yield this.removeRelations();
            return true;
        });
    }
    /**
     * @private
     */
    removeRelations() {
        const { rels } = getProtoData(this, false);
        return Promise.all(rels.map(rel => this.removeEntityRelation(rel))).then(() => {
            return;
        });
    }
    getProp(key) {
        return this[key];
    }
    getRelationDescriptor(keyName) {
        const { rels } = getProtoData(this);
        const rel = rels.find(rel => {
            return rel.key === keyName;
        });
        return rel;
    }
    getRelationEntity(keyName) {
        const rel = this.getRelationDescriptor(keyName);
        const { factory } = getSlothData(this);
        if ('belongsTo' in rel) {
            return rel.belongsTo().findById(factory, this.getProp(keyName));
        }
        throw new Error(`Cannot use getRelationEntity on ${keyName} as hasMany relation`);
    }
    removeEntityRelation(rel) {
        return __awaiter(this, void 0, void 0, function* () {
            const { factory, name } = getSlothData(this);
            if ('belongsTo' in rel && rel.cascade) {
                const relId = this.getProp(rel.key);
                const children = yield factory(name).allDocs({
                    include_docs: false,
                    startkey: join(relId, '/'),
                    endkey: join(relId, '/\uffff')
                });
                if (children.rows.length === 0) {
                    try {
                        const parent = yield this.getRelationEntity(rel.key);
                        yield parent.remove();
                    }
                    catch (err) {
                        if (err.message === 'missing') {
                            return;
                        }
                        throw err;
                    }
                }
            }
            if ('hasMany' in rel && rel.cascade !== false) {
                const db = rel.hasMany().withRoot(this._id);
                const children = yield db.findAllDocs(factory);
                yield Promise.all(children.map(child => child.remove()));
            }
        });
    }
}

/**
 * FSA Compliant changes
 */
/**
 * Action types, describes the type of event
 * (added, removed or changed)
 */
var ActionType;
(function (ActionType) {
    ActionType["ADDED"] = "@@DOCUMENT_ADDED";
    ActionType["REMOVED"] = "@@DOCUMENT_REMOVED";
    ActionType["CHANGED"] = "@@DOCUMENT_CHANGED";
})(ActionType || (ActionType = {}));

function belongsToMapper(target, keyName) {
    return (factory) => {
        const { rels } = getProtoData(target);
        const rel = rels.find(({ key }) => key === keyName);
        if (!rel) {
            throw new Error(`No relation available for ${keyName}`);
        }
        if ('belongsTo' in rel) {
            return rel.belongsTo().findById(factory, target[keyName]);
        }
        throw new Error('Unsupported relation');
    };
}

const debug$3 = src('slothdb');
/**
 * This represent a Database
 *
 * @typeparam S the database schema
 * @typeparam E the Entity
 * @typeparam V the (optional) view type that defines a list of possible view IDs
 */
class SlothDatabase {
    /**
     * Create a new database instance
     * @param factory the pouch factory to use
     * @param name the database name
     * @param model the model constructor
     * @param root the root name, which is the startKey. I don't recommend it
     */
    constructor(model, root = '') {
        this._subscribers = [];
        this._model = model;
        this._root = root;
        const { name } = getProtoData(model.prototype);
        /* istanbul ignore if  */
        if (!name) {
            throw new Error('SlothEntity decorator is required');
        }
        this._name = name;
    }
    /**
     * Join URI params provided as specified by SlothURI
     * Useful to recreate document id from URL params
     * Please note that
     * @param props the props
     * @param field
     */
    joinURIParams(props, field = '_id') {
        const idURI = getProtoData(this._model.prototype).uris.find(({ name }) => name === field);
        if (!idURI) {
            throw new Error(`Field ${field} not found in URIs`);
        }
        return join(idURI.prefix, ...idURI.propsKeys.map(key => props[key]));
    }
    /**
     * Run a query
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     * @param includeDocs include_docs
     */
    query(factory, view, startKey, endKey = typeof startKey === 'string'
        ? join(startKey, '\uffff')
        : undefined, includeDocs = false) {
        return factory(this._name)
            .query(view, {
            startkey: startKey,
            endkey: endKey,
            include_docs: includeDocs
        })
            .catch(err => {
            if (err.name === 'not_found') {
                debug$3(`Design document '%s' is missing, generating views...`, view);
                /* istanbul ignore if  */
                if (this._setupPromise) {
                    this._setupPromise.then(() => {
                        return this.query(factory, view, startKey, endKey, includeDocs);
                    });
                }
                this._setupPromise = this.initSetup(factory);
                return this._setupPromise.then(() => {
                    debug$3('Created design documents');
                    this._setupPromise = undefined;
                    return this.query(factory, view, startKey, endKey, includeDocs);
                });
            }
            throw err;
        });
    }
    /**
     * Queries and maps docs to Entity objects
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryDocs(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, true).then(({ rows }) => {
            return rows.map(({ doc }) => new this._model(factory, doc));
        });
    }
    /**
     * Queries keys. Returns an array of emitted keys
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeys(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, false).then(({ rows }) => {
            return rows.map(({ key }) => key);
        });
    }
    /**
     * Queries keys/_id map. Returns a map of emitted keys/ID
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeysIDs(factory, view, startKey, endKey) {
        return this.query(factory, view, startKey, endKey, false).then(({ rows }) => {
            return rows.reduce((acc, { key, id }) => (Object.assign({}, acc, { [key]: id })), {});
        });
    }
    /**
     * Returns a database that will only find entities with _id
     * starting with the root path
     * @param root the root path
     */
    withRoot(root) {
        return new SlothDatabase(this._model, join(this._root, root));
    }
    /**
     * Fetches all documents IDs for this database and return them
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of string IDs
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllIDs(factory, startKey = this._root || '', endKey = join(startKey, '\uffff')) {
        const db = factory(this._name);
        return db
            .allDocs({
            include_docs: false,
            startkey: startKey,
            endkey: endKey
        })
            .then(({ rows }) => {
            return rows.map(({ id }) => id);
        });
    }
    /**
     * Fetches all documents for this database and map them with the model
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of entity instances
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllDocs(factory, startKey = this._root || '', endKey = join(startKey, '\uffff')) {
        const db = factory(this._name);
        return db
            .allDocs({
            include_docs: true,
            startkey: startKey,
            endkey: endKey
        })
            .then(({ rows }) => {
            return rows.map(({ doc }) => this.create(factory, doc));
        });
    }
    /**
     * Fetch a docuemt from the database
     * @param factory the PouchDB factory to use
     * @param id the document identifier to fetch
     * @return a promise resolving with the entity instance
     */
    findById(factory, id) {
        return factory(this._name)
            .get(id)
            .then(res => {
            return new this._model(factory, res);
        });
    }
    /**
     * Create a new model instance
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    create(factory, props) {
        return new this._model(factory, props);
    }
    /**
     * Create a new model instance and save it to database
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    put(factory, props) {
        const doc = new this._model(factory, props);
        return doc.save().then(() => doc);
    }
    /**
     * Subscribes a function to PouchDB changes, so that
     * the function will be called when changes are made
     *
     * @param factory the PouchDB factory
     * @param sub the subscriber function
     * @see [[Subscriber]]
     * @see [[ChangeAction]]
     */
    subscribe(factory, sub) {
        if (!this.getSubscriberFor(factory)) {
            debug$3('Creating changes ');
            const changes = factory(this._name)
                .changes({
                since: 'now',
                live: true,
                include_docs: true
            })
                .on('change', ({ deleted, doc, id }) => {
                if (deleted || !doc) {
                    return this.dispatch(factory, {
                        type: ActionType.REMOVED,
                        payload: { [this._name]: id },
                        meta: {}
                    });
                }
                if (doc._rev.match(/^1-/)) {
                    return this.dispatch(factory, {
                        type: ActionType.ADDED,
                        payload: { [this._name]: doc },
                        meta: { revision: doc._rev }
                    });
                }
                return this.dispatch(factory, {
                    type: ActionType.CHANGED,
                    payload: { [this._name]: doc },
                    meta: { revision: doc._rev }
                });
            });
            this._subscribers.push({ factory, sub, changes });
            return;
        }
        const { changes } = this.getSubscriberFor(factory);
        this._subscribers.push({ factory, sub, changes });
    }
    /**
     * Unsubscribe a subscriber, so it will not be called anymore
     * Possibly cancel PouchDB changes
     *
     * @param factory The pouchDB factory to unsubscribe from
     * @param sub the subscriber to unsubscribe
     */
    cancel(factory, sub) {
        const index = this._subscribers.findIndex(el => el.factory === factory && el.sub === sub);
        const [{ changes }] = this._subscribers.splice(index, 1);
        if (!this.getSubscriberFor(factory)) {
            changes.cancel();
        }
    }
    /**
     * Creates view documents (if required)
     * @param factory
     */
    initSetup(factory) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.setupViews(factory);
        });
    }
    getSubscriberFor(factory) {
        return this._subscribers.find(el => el.factory === factory);
    }
    dispatch(facto, action) {
        this._subscribers.forEach(({ sub, factory }) => facto === factory && sub(action));
    }
    setupViews(factory) {
        const { views } = getProtoData(this._model.prototype);
        const db = factory(this._name);
        const promises = views.map(({ name, id, code }) => () => __awaiter(this, void 0, void 0, function* () {
            const views = {};
            let _rev;
            try {
                const doc = (yield db.get(`_design/${id}`));
                if (doc.views[name] && doc.views[name].map === code) {
                    // view already exists and is up-to-date
                    return;
                }
                Object.assign(views, doc.views);
                _rev = doc._rev;
            }
            catch (err) {
                // Do nothing
            }
            yield db.put(Object.assign({}, {
                _id: `_design/${id}`,
                views: Object.assign({}, views, { [name]: {
                        map: code
                    } })
            }, _rev ? { _rev } : {}));
        }));
        return promises.reduce((acc, fn) => {
            return acc.then(() => fn());
        }, Promise.resolve());
    }
}

/**
 * Creates a view for a field. This function does not modify the
 * behavior of the current field, hence requires another decorator
 * such as SlothURI or SlothField.  The view  will be created by the SlothDatabase
 *
 * @export
 * @template S
 * @param {(doc: S, emit: Function) => void} fn the view function, as arrow or es5 function
 * @param {string} [docId='views'] the _design document identifier
 * @param {string} [viewId] the view identifier, default by_<property name>
 * @returns the decorator to apply on the field
 */
function SlothView(fn, viewId, docId = 'views') {
    return (target, key) => {
        const desc = Reflect.getOwnPropertyDescriptor(target, key);
        if (desc) {
            if (!desc.get && !desc.set) {
                throw new Error('Required SlothView on top of another decorator');
            }
        }
        const fun = `function (__doc) { 
      (${fn.toString()})(__doc, emit);
    }`;
        const { views, fields } = getProtoData(target, true);
        const field = fields.find(field => field.key === key);
        if (!field) {
            throw new Error('Required SlothView on top of a SlothField');
        }
        views.push({
            id: docId,
            name: viewId || `by_${field.docKey}`,
            function: fn,
            code: fun
        });
    };
}

/**
 * Creates an index for a field. It's a view function that simply emits
 * the document key
 *
 * @see [[SlothDatabase.queryDocs]]
 * @export
 * @template S
 * @param {(doc: S, emit: Function) => void} fn the view function, as arrow or es5 function
 * @param {string} [docId='views'] the _design document identifier
 * @param {string} [viewId] the view identifier, default by_<property name>
 * @returns the decorator to apply on the field
 */
function SlothIndex(viewId, docId) {
    return (target, key) => {
        const field = getProtoData(target).fields.find(field => field.key === key);
        if (!field) {
            throw new Error('Please use SlothIndex on top of a SlothField');
        }
        SlothView(new Function('doc', 'emit', `emit(doc['${field.docKey}'].toString());`), viewId, docId)(target, key);
    };
}

export { SlothEntity, SlothURI, SlothRel, SlothField, BaseEntity, ActionType, SlothDatabase, SlothView, SlothIndex, belongsToMapper };
