/// <reference types="pouchdb-mapreduce" />
/// <reference types="pouchdb-core" />
import PouchFactory from './PouchFactory';
import BaseEntity from './BaseEntity';
import { Subscriber, ChangeAction } from './changes';
import EntityConstructor from '../helpers/EntityConstructor';
import Dict from '../helpers/Dict';
/**
 * This represent a Database
 *
 * @typeparam S the database schema
 * @typeparam E the Entity
 * @typeparam V the (optional) view type that defines a list of possible view IDs
 */
export default class SlothDatabase<S extends {
    _id: string;
}, E extends BaseEntity<S>, V extends string = never> {
    _root: string;
    /**
     *
     * @private
     * @type {string}
     * @memberof SlothDatabase
     */
    _name: string;
    /**
     *
     *
     * @type {T}
     * @memberof SlothDatabase
     * @private
     */
    _model: EntityConstructor<S, E>;
    private _subscribers;
    private _setupPromise?;
    /**
     * Create a new database instance
     * @param factory the pouch factory to use
     * @param name the database name
     * @param model the model constructor
     * @param root the root name, which is the startKey. I don't recommend it
     */
    constructor(model: EntityConstructor<S, E>, root?: string);
    /**
     * Join URI params provided as specified by SlothURI
     * Useful to recreate document id from URL params
     * Please note that
     * @param props the props
     * @param field
     */
    joinURIParams(props: Partial<S>, field?: string): string;
    /**
     * Run a query
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     * @param includeDocs include_docs
     */
    query(factory: PouchFactory<S>, view: V, startKey?: any, endKey?: any, includeDocs?: boolean): Promise<PouchDB.Query.Response<S>>;
    /**
     * Queries and maps docs to Entity objects
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryDocs(factory: PouchFactory<S>, view: V, startKey?: any, endKey?: any): Promise<E[]>;
    /**
     * Queries keys. Returns an array of emitted keys
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeys(factory: PouchFactory<S>, view: V, startKey?: any, endKey?: any): Promise<string[]>;
    /**
     * Queries keys/_id map. Returns a map of emitted keys/ID
     *
     * @param factory the pouch factory
     * @param view the view identifier
     * @param startKey the optional startkey
     * @param endKey the optional endkey
     */
    queryKeysIDs(factory: PouchFactory<S>, view: V, startKey?: any, endKey?: any): Promise<Dict<string>>;
    /**
     * Returns a database that will only find entities with _id
     * starting with the root path
     * @param root the root path
     */
    withRoot(root: string): SlothDatabase<S, E, never>;
    /**
     * Fetches all documents IDs for this database and return them
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of string IDs
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllIDs(factory: PouchFactory<S>, startKey?: string, endKey?: string): Promise<string[]>;
    /**
     * Fetches all documents for this database and map them with the model
     *
     * @param {PouchFactory<S>} factory the PouchDB factory to use
     * @param {string} [startKey=''] the startkey to use
     * @param {string} [endKey=path.join(startKey, '\uffff')] the endkey to use
     * @returns a promise that resolves into an array of entity instances
     * @see PouchDB#allDocs
     * @memberof SlothDatabase
     */
    findAllDocs(factory: PouchFactory<S>, startKey?: string, endKey?: string): Promise<E[]>;
    /**
     * Fetch a docuemt from the database
     * @param factory the PouchDB factory to use
     * @param id the document identifier to fetch
     * @return a promise resolving with the entity instance
     */
    findById(factory: PouchFactory<S>, id: string): Promise<E>;
    /**
     * Create a new model instance
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    create(factory: PouchFactory<S>, props: Partial<S>): E;
    /**
     * Create a new model instance and save it to database
     * @param factory The database factory to attach to the model
     * @param props the entity properties
     * @returns an entity instance
     */
    put(factory: PouchFactory<S>, props: Partial<S>): Promise<E>;
    /**
     * Subscribes a function to PouchDB changes, so that
     * the function will be called when changes are made
     *
     * @param factory the PouchDB factory
     * @param sub the subscriber function
     * @see [[Subscriber]]
     * @see [[ChangeAction]]
     */
    subscribe(factory: PouchFactory<S>, sub: Subscriber<S>): void;
    /**
     * Unsubscribe a subscriber, so it will not be called anymore
     * Possibly cancel PouchDB changes
     *
     * @param factory The pouchDB factory to unsubscribe from
     * @param sub the subscriber to unsubscribe
     */
    cancel(factory: PouchFactory<S>, sub: Subscriber<S>): void;
    /**
     * Creates view documents (if required)
     * @param factory
     */
    initSetup(factory: PouchFactory<S>): Promise<void>;
    protected getSubscriberFor(factory: PouchFactory<S>): {
        factory: any;
        sub: Subscriber<S>;
        changes: PouchDB.Core.Changes<S>;
    } | undefined;
    protected dispatch(facto: PouchFactory<any>, action: ChangeAction<S>): void;
    private setupViews(factory);
}
